import { Service } from "../service"
import { UserDefinitionService, User } from "./userDefinitionService"
import { MongodbService } from "./mongodbService"
import * as mongodb from "mongodb"
import * as crypto from "crypto"
import * as Leaf from "leaf-ts"
export type Callback<T> = (err?: Error, result?: T) => void

export const Errors = Leaf.ErrorDoc.create()
    .define("UnknownError")
    .define("DBError")
    .define("AlreadyExists")
    .define("UsernameAlreadyExists")
    .define("EmailAlreadyExists")
    .define("InvalidParameter")
    .generate() as Leaf.ErrorsOf<{
        UnknownError: {}
        DBError: {}
        AlreadyExists: {}
        InvalidParameter: {}
        UsernameAlreadyExists: {}
        EmailAlreadyExists: {}
    }>
function clearUserPassword(obj: any, name: string, descriptor: TypedPropertyDescriptor<Function>) {
    let fn = descriptor.value
    descriptor.value = function (option, callback: Function) {
        fn.call(this, option, (err, value) => {
            if (value instanceof Array) {
                value = value.forEach((item) => {
                    delete item.passwordHash
                })
            } else if (value) {
                delete value.passwordHash
            }
            callback(err, value)
        })
    }
}
export class UserService<TUser extends User = User> extends Service {
    readonly name = "UserService"
    readonly dependencies = ["MongodbService", "UserDefinitionService"] as any
    readonly collectionName = "user"
    public readonly Errors = Errors
    constructor(config: {
        user?: {
            salt?: string
            blacklist?: string[]
        }
    } = {}) {
        super()
        if (config.user) {
            if (config.user.salt) {
                this.salt = config.user.salt
            }
            if (config.user.blacklist) {
                this.blacklist = config.user.blacklist
            }
        }
    }
    private blacklist: string[] = []
    private salt: string = "default-user-secret-salt"
    services: {
        MongodbService: MongodbService
        UserDefinitionService: UserDefinitionService<TUser>
    }
    userCollection: mongodb.Collection
    initialize() {
        this.salt
        this.userCollection = this.services.MongodbService.db.collection(this.collectionName)
        this.userCollection.createIndex({
            username: 1
        }, {
                unique: true
            })
        this.userCollection.createIndex({
            email: 1
        }, {
                unique: true
            })
        this.userCollection.createIndex({
            passwordHash: 1
        })
    }
    private createPasswordHashString(rawPassword: string): string {
        let level1 = crypto.createHash("sha256").update(rawPassword + this.salt).digest('hex')
        let level2 = crypto.createHash("sha512").update(level1 + this.salt).digest('hex')
        return level2
    }
    private translateMongoId(doc: { _id: mongodb.ObjectID } & TUser): TUser {
        if (!doc) return null
        let id = doc._id
        delete doc._id
        doc.id = id.toString()
        return doc
    }
    public toPublicUser(user: TUser): TUser {
        delete user.passwordHash
        delete user.password
        return user
    }
    @clearUserPassword
    getUserById(id: string, callback: Callback<TUser>) {
        let _id = new mongodb.ObjectID(id)
        this.userCollection.findOne({ _id }, (error: Error, doc: { _id: mongodb.ObjectID } & TUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let user = this.translateMongoId(doc)
            callback(null, user)
        })
    }
    @clearUserPassword
    getUserByUsername(username: string, callback: Callback<TUser>) {
        username = username.toLowerCase()
        this.userCollection.findOne({ username }, (error: Error, doc: { _id: mongodb.ObjectID } & TUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let user = this.translateMongoId(doc)
            callback(null, user)
        })
    }
    @clearUserPassword
    getUserByEmail(email: string, callback: Callback<TUser>) {
        email = email.toLowerCase()
        this.userCollection.findOne({ email }, (error: Error, doc: { _id: mongodb.ObjectID } & TUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let user = this.translateMongoId(doc)
            callback(null, user)
        })
    }
    @clearUserPassword
    getUserByLogin(option: { login: string, password: string }, callback: Callback<TUser>) {
        let login = option.login.toLowerCase()
        let passwordHash = this.createPasswordHashString(option.password)
        this.userCollection.findOne({ passwordHash: passwordHash, $or: [{ username: login }, { email: login }] }, (error: Error, doc: { _id: mongodb.ObjectID } & TUser) => {
            if (error) {
                callback(new Errors.DBError("DatabaseError", { via: error }))
                return
            }
            let user = this.translateMongoId(doc)
            callback(null, user)
        })
    }

    @clearUserPassword
    createUser(partialUser: Partial<TUser>, callback: Callback<TUser>) {
        if (this.blacklist.indexOf(partialUser.username) >= 0) {
            callback(new Errors.InvalidParameter("Invalid username"))
            return
        }
        if (!partialUser.username || !partialUser.email) {
            callback(new Errors.InvalidParameter("Empty username or email"))
            return
        }
        // use username as display name by default
        if (!partialUser.displayName) {
            partialUser.displayName = partialUser.username
        }
        // then convert username to all lowercase
        partialUser.username = partialUser.username.toLocaleLowerCase()
        partialUser.email = partialUser.email.toLowerCase()
        try {
        }
        catch (error) {
            callback(error)
            return
        }
        try {
            this.services.UserDefinitionService.UserValidator.check(partialUser)
        } catch (e) {
            callback(new Errors.InvalidParameter())
            return
        }
        let passwordHash = this.createPasswordHashString(partialUser.password)
        let user = {
            username: partialUser.username,
            displayName: partialUser.displayName,
            passwordHash: passwordHash,
            email: partialUser.email,
            crteateAt: Date.now()
        }
        this.getUserByUsername(user.username, (error: Error, doc: TUser) => {
            if (doc) {
                callback(new Errors.UsernameAlreadyExists())
                return
            }
            this.getUserByEmail(user.email, (error: Error, doc: TUser) => {
                if (doc) {
                    callback(new Errors.EmailAlreadyExists())
                    return
                }
                this.userCollection.insert(user, (error: Error, doc: mongodb.InsertOneWriteOpResult) => {
                    if (error) {
                        // on a edge case that 2 user try to use a same username and something like that
                        callback(new Errors.DBError("DatabaseError", { via: error }))
                        return
                    }
                    let user = this.translateMongoId(doc.ops[0])
                    callback(null, user)
                })
            })
        })
    }
    @clearUserPassword
    updateUserPasswordByUsername(option: { username: string, password: string }, callback: Callback<TUser>) {
        try {
            this.services.UserDefinitionService.UserValidator.check({ password: option.password })
        }
        catch (e) {
            callback(new Errors.InvalidParameter())
            return
        }
        const username = option.username.toLowerCase()
        let passwordHash = this.createPasswordHashString(option.password)
        this.userCollection.findOneAndUpdate({ username }, { $set: { passwordHash } }, { returnOriginal: false }, (error, doc: mongodb.FindAndModifyWriteOpResultObject) => {
            if (error) {
                callback(error)
                return
            }
            let user = this.translateMongoId(doc.value)
            callback(null, user)
        })
    }
    @clearUserPassword
    updateUserPasswordById(option: { id: string, password: string }, callback: Callback<TUser>) {
        let _id = new mongodb.ObjectID(option.id)
        try {
            this.services.UserDefinitionService.UserValidator.check({ password: option.password })
        }
        catch (e) {
            callback(new Errors.InvalidParameter())
            return
        }
        let passwordHash = this.createPasswordHashString(option.password)
        this.userCollection.findOneAndUpdate({ _id }, { $set: { passwordHash } }, { returnOriginal: false }, (error, doc: mongodb.FindAndModifyWriteOpResultObject) => {
            if (error) {
                callback(error)
                return
            }
            let user = this.translateMongoId(doc.value)
            callback(null, user)
        })
    }
}
