import { Service } from "./service"
import * as express from "express"
import * as urlModule from "url"
import * as colors from "colors"
import * as PathToRegExp from "path-to-regexp"
let useColors = colors

export interface Request extends express.Request {
}
export interface Response extends express.Response { }

export abstract class RouteService extends Service {
    public routes: RouteConstructor[] = []
    public prefix: string = "/"
    addRoute(RouteConstructor) {
        this.routes.push(RouteConstructor)
    }
    installTo(server: express.Application) {
        for (let cons of this.routes) {
            let route = new cons(this)
            if (route instanceof MiddlewareRoute) {
                let handler = (req, res, next) => {
                    let route = new cons(this)
                    route.services = this.services
                    route.handleRawRequest(req, res, next)
                }
                if (route.path) {
                    server.use(route.path, handler)
                } else {
                    server.use(handler)
                }
                console.log(`Install middleware ${useColors.yellow(cons.name)}`)
                continue
            }
            let path = urlModule.resolve(this.prefix, route.path)
            let method = route.method.toLowerCase()
            if (route.name) {
                console.log(`Install ${useColors.yellow(route.name)} ~ ${useColors.yellow(route.method)}~${useColors.yellow(path)}`)
            } else {
                console.log(`Install ${useColors.yellow(route.method)} => ${useColors.yellow(path)}`)
            }
            server[method](path, (req, res, next) => {
                let route = new cons(this)
                route.services = this.services
                route.handleRawRequest(req, res, next)
            })

        }
    }
    addRoutes(...obj: any[]) {
        for (let item of obj) {
            if (this.isRouteConstructor(item)) {
                this.addRoute(item)
            } else {
                for (let name in item) {
                    let target = item[name]
                    if (this.isRouteConstructor(target)) {
                        this.addRoute(target)
                    }
                }
            }
        }
    }
    isRouteConstructor(obj) {
        return obj && obj.prototype && obj.prototype["thisObjectIsTruelyARoute"] == "thisObjectIsTruelyARoute"
    }
}
export interface RouteConstructor {
    new (service?: RouteService): Route<any>
}
export abstract class Route<TMIME extends MIME> {
    constructor(public service: RouteService) {
    }
    name: string
    mime: TMIME
    abstract path: string
    method: METHOD_ENUM = "GET"
    params = {}
    body = {}
    info: any
    services: {
        [index: string]: Service
    } | any
    query = {}
    next: Function
    req: Request
    res: Response
    handleRawRequest(req, res, next) {
        if (!req._routeInfo) {
            req._routeInfo = {}
        }
        this.req = req
        this.res = res
        this.info = req._routeInfo
        this.next = next
        this.body = req.body || {}
        this.params = req.params || {}
        this.query = req.query || {}
        this.handle()
    }
    end(content: any) {
        this.res.contentType(this.mime)
        this.res.end(content)
    }
    abstract handle();
}
Route.prototype["thisObjectIsTruelyARoute"] = "thisObjectIsTruelyARoute"
export abstract class DynamicResourceRoute<T extends MIME> extends Route<T>{ }
export abstract class MiddlewareRoute extends Route<never>{
    path = null
    method: Method.ALL
    mock(url: string, data: {
        [name: string]: any
    }, callback: (err?: any) => void) {
        for (let name in data) {
            this[name] = data[name]
        }
        let keys = []
        if (this.path) {
            let reg = PathToRegExp(this.path, keys, {})
            let match = reg.exec(url)
            if (!match) {
                callback(new Error(`${url} doesn't match ${this.path}`))
                return
            }
            let params = {}
            for (let i = 0; i < keys.length; i++) {
                let k = keys[i]
                params[k.name] = match[i + 1]
            }
            this.params = params
            if (!reg.test(url)) {
                callback(new Error("Url not match"))
                return
            }
        }
        this.next = () => {
            callback()
        }
        this.handle()
    }
}
export abstract class PageRoute extends Route<"text/html">{
    method: Method.GET
    readonly mime = "text/html"
    html(content: string) {
        this.end(content)
    }
}
export abstract class APIRoute<TResponse> extends Route<"application/json"> {
    done(err: Error, result: TResponse) {
        if (err) {
            this.error(err)
        } else {
            this.success(result)
        }
    }
    success(data: TResponse) {
        if (this._success) {
            this._success(data)
            return
        }
        this.json({
            state: true,
            data
        })
    }
    error(err: Error) {
        if (this._error) {
            this._error(err)
            return
        }
        this.json({
            state: false,
            error: err
        })
    }
    // for mockup
    _error: (err: Error) => void
    _success: (data: any) => void
    json(data) {
        this.res.json(data)
    }
    mock(url: string, data: {
        [name: string]: any
    }, callback: (err?: any, data?: TResponse) => void) {
        for (let name in data) {
            this[name] = data[name]
        }
        let keys = []
        let reg = PathToRegExp(this.path, keys, {})
        let match = reg.exec(url)
        if (!match) {
            callback(new Error(`${url} doesn't match ${this.path}`))
            return
        }
        let params = {}
        for (let i = 0; i < keys.length; i++) {
            let k = keys[i]
            params[k.name] = match[i + 1]
        }
        this.params = params
        if (!reg.test(url)) {
            callback(new Error("Url not match"))
            return
        }
        let returnError: Error = null
        let returnValue: TResponse = null
        this._error = (err) => {
            clear()
            callback(err, null)
        }
        this._success = (data) => {
            clear()
            callback(null, data)
        }
        let clear = () => {
            this._error = null
            this._success = null
        }
        this.handle()
    }
}
export type MIME = "text/html" | "text/javascript" | "text/json" | "application/json" | string
export type METHOD_ENUM = "GET" | "POST" | "DELETE" | "PUT" | "ALL"
export type Method<T extends METHOD_ENUM> = string & T
export namespace Method {
    export const GET: Method<"GET"> = "GET"
    export const POST: Method<"POST"> = "POST"
    export const DELETE: Method<"DELETE"> = "DELETE"
    export const PUT: Method<"PUT"> = "PUT"
    export const ALL: Method<"ALL"> = "ALL"
    export type POST = Method<"POST">
    export type GET = Method<"GET">
    export type DELETE = Method<"DELETE">
    export type PUT = Method<"PUT">
    export type ALL = Method<"ALL">
}
